@@ -70,6 +70,8 @@ def convertsink(ui, path, type):
                 return sink(ui, path)
         except NoRepo, inst:
             ui.note(_("convert: %s\n") % inst)
+        except MissingTool, inst:
+            raise util.Abort('%s\n' % inst)
     raise util.Abort(_('%s: unknown repository type') % path)
 
 class progresssource(object):
@@ -81,7 +83,7 @@ class progresssource(object):
 
     def getfile(self, file, rev):
         self.retrieved += 1
-        self.ui.progress(_('retrieving file'), self.retrieved,
+        self.ui.progress(_('getting files'), self.retrieved,
                          item=file, total=self.filecount)
         return self.source.getfile(file, rev)
 
@@ -89,7 +91,7 @@ class progresssource(object):
         return self.source.lookuprev(rev)
 
     def close(self):
-        self.ui.progress(_('retrieving file'), None)
+        self.ui.progress(_('getting files'), None)
 
 class converter(object):
     def __init__(self, ui, source, dest, revmapfile, opts):
@@ -112,8 +114,8 @@ class converter(object):
         if authorfile and os.path.exists(authorfile):
             self.readauthormap(authorfile)
         # Extend/Override with new author map if necessary
-        if opts.get('authors'):
-            self.readauthormap(opts.get('authors'))
+        if opts.get('authormap'):
+            self.readauthormap(opts.get('authormap'))
             self.authorfile = self.dest.authorfile()
 
         self.splicemap = mapfile(ui, opts.get('splicemap'))
@@ -344,14 +346,14 @@ class converter(object):
             c = None
 
             self.ui.status(_("converting...\n"))
-            for i, c in enumerate(t):                
+            for i, c in enumerate(t):
                 num -= 1
                 desc = self.commitcache[c].desc
                 if "\n" in desc:
                     desc = desc.splitlines()[0]
                 # convert log message to local encoding without using
-                # tolocal() because encoding.encoding conver() use it as
-                # 'utf-8'
+                # tolocal() because the encoding.encoding convert()
+                # uses is 'utf-8'
                 self.ui.status("%d %s\n" % (num, recode(desc)))
                 self.ui.note(_("source: %s\n") % recode(c))
                 self.ui.progress(_('converting'), i, unit=_('revisions'),
@@ -376,6 +378,16 @@ class converter(object):
                     if tagsparents:
                         self.map[tagsparents[0][0]] = nrev
 
+            bookmarks = self.source.getbookmarks()
+            cbookmarks = {}
+            for k in bookmarks:
+                v = bookmarks[k]
+                if self.map.get(v, SKIPREV) != SKIPREV:
+                    cbookmarks[k] = self.map[v]
+
+            if c and cbookmarks:
+                self.dest.putbookmarks(cbookmarks)
+
             self.writeauthormap()
         finally:
             self.cleanup()
@@ -392,6 +404,10 @@ def convert(ui, src, dest=None, revmapfile=None, **opts):
     orig_encoding = encoding.encoding
     encoding.encoding = 'UTF-8'
 
+    # support --authors as an alias for --authormap
+    if not opts.get('authormap'):
+        opts['authormap'] = opts.get('authors')
+
     if not dest:
         dest = hg.defaultdest(src) + "-hg"
         ui.status(_("assuming destination %s\n") % dest)
