@@ -41,7 +41,7 @@ def _updatesample(dag, nodes, sample, always, quicksamplesize=0):
 def _setupsample(dag, nodes, size):
     if len(nodes) <= size:
         return set(nodes), None, 0
-    always = set(dag.heads())
+    always = dag.headsetofconnecteds(nodes)
     desiredlen = size - len(always)
     if desiredlen <= 0:
         # This could be bad if there are very many heads, all unknown to the
@@ -124,7 +124,7 @@ def findcommonheads(ui, local, remote,
 
     srvheads = dag.internalizeall(srvheadhashes, filterunknown=True)
     if len(srvheads) == len(srvheadhashes):
-        ui.note("all remote heads known locally\n")
+        ui.debug("all remote heads known locally\n")
         return (srvheadhashes, False, srvheadhashes,)
 
     if sample and util.all(yesno):
@@ -190,4 +190,5 @@ def findcommonheads(ui, local, remote,
             ui.warn(_("warning: repository is unrelated\n"))
         return (set([nullid]), True, srvheadhashes,)
 
-    return (dag.externalizeall(result), True, srvheadhashes,)
+    anyincoming = (srvheadhashes != [nullid])
+    return dag.externalizeall(result), anyincoming, srvheadhashes
